# THIS FILE HAS BEEN AUTOGENERATED. To update:
# 1. modify the `_deps` dict in setup.py
# 2. run `make fix-repo``
deps = {
    "Pillow": "Pillow>=10.0.1,<=15.0",
    "accelerate": "accelerate>=1.1.0",
    "av": "av",
    "beautifulsoup4": "beautifulsoup4",
    "blobfile": "blobfile",
    "codecarbon": "codecarbon>=2.8.1",
    "datasets": "datasets>=2.15.0",
    "deepspeed": "deepspeed>=0.9.3",
    "dill": "dill<0.3.5",
    "evaluate": "evaluate>=0.4.6",
    "faiss-cpu": "faiss-cpu",
    "fastapi": "fastapi",
    "filelock": "filelock",
    "fugashi": "fugashi>=1.0",
    "GitPython": "GitPython<3.1.19",
    "hf-doc-builder": "hf-doc-builder>=0.3.0",
    "huggingface-hub": "huggingface-hub>=1.3.0,<2.0",
    "ipadic": "ipadic>=1.0.0,<2.0",
    "jinja2": "jinja2>=3.1.0",
    "jmespath": "jmespath>=1.0.1",
    "kenlm": "kenlm",
    "kernels": "kernels>=0.10.2,<0.11",
    "librosa": "librosa",
    "mistral-common[image]": "mistral-common[image]>=1.8.8",
    "nltk": "nltk<=3.8.1",
    "num2words": "num2words",
    "numpy": "numpy>=1.17",
    "openai": "openai>=1.98.0",
    "opencv-python": "opencv-python",
    "optimum-benchmark": "optimum-benchmark>=0.3.0",
    "optuna": "optuna",
    "pandas": "pandas<2.3.0",
    "packaging": "packaging>=20.0",
    "parameterized": "parameterized>=0.9",
    "peft": "peft>=0.18.0",
    "phonemizer": "phonemizer",
    "protobuf": "protobuf",
    "psutil": "psutil",
    "pyyaml": "pyyaml>=5.1",
    "pydantic": "pydantic>=2",
    "pytest": "pytest>=7.2.0,<9.0.0",
    "pytest-asyncio": "pytest-asyncio>=1.2.0",
    "pytest-random-order": "pytest-random-order",
    "pytest-rerunfailures": "pytest-rerunfailures<16.0",
    "pytest-timeout": "pytest-timeout",
    "pytest-env": "pytest-env",
    "pytest-xdist": "pytest-xdist",
    "pytest-order": "pytest-order",
    "python": "python>=3.10.0",
    "regex": "regex!=2019.12.17",
    "rhoknp": "rhoknp>=1.1.0,<1.3.1",
    "rjieba": "rjieba",
    "rouge-score": "rouge-score!=0.0.7,!=0.0.8,!=0.1,!=0.1.1",
    "ruff": "ruff==0.14.10",
    "ty": "ty==0.0.12",
    "sacrebleu": "sacrebleu>=1.4.12,<2.0.0",
    "sacremoses": "sacremoses",
    "safetensors": "safetensors>=0.4.3",
    "sagemaker": "sagemaker>=2.31.0",
    "schedulefree": "schedulefree>=1.2.6",
    "scikit-learn": "scikit-learn",
    "scipy": "scipy",
    "sentencepiece": "sentencepiece>=0.1.91,!=0.1.92",
    "starlette": "starlette",
    "sudachipy": "sudachipy>=0.6.6",
    "sudachidict_core": "sudachidict_core>=20220729",
    "tensorboard": "tensorboard",
    "timeout-decorator": "timeout-decorator",
    "tiktoken": "tiktoken",
    "timm": "timm>=1.0.23",
    "tokenizers": "tokenizers>=0.22.0,<=0.23.0",
    "torch": "torch>=2.4",
    "torchaudio": "torchaudio",
    "torchvision": "torchvision",
    "pyctcdecode": "pyctcdecode>=0.4.0",
    "tqdm": "tqdm>=4.27",
    "typer": "typer",
    "unidic": "unidic>=1.0.2",
    "unidic_lite": "unidic_lite>=1.0.7",
    "urllib3": "urllib3<2.0.0",
    "uvicorn": "uvicorn",
    "pytest-rich": "pytest-rich",
    "libcst": "libcst",
    "rich": "rich",
    "ray[tune]": "ray[tune]>=2.7.0",
    "opentelemetry-api": "opentelemetry-api",
    "opentelemetry-exporter-otlp": "opentelemetry-exporter-otlp",
    "opentelemetry-sdk": "opentelemetry-sdk",
}
